##################################################################################
# Copyright (c) 2011, Daniel Urieli, Peter Stone
# University of Texas at Austin
# All right reserved
# 
# Based On:
# 
# Copyright (c) 2000-2003, Jelle Kok, University of Amsterdam
# All rights reserved.
# 
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
# 
# 1. Redistributions of source code must retain the above copyright notice, this
# list of conditions and the following disclaimer.
# 
# 2. Redistributions in binary form must reproduce the above copyright notice,
# this list of conditions and the following disclaimer in the documentation
# and/or other materials provided with the distribution.
# 
# 3. Neither the name of the University of Amsterdam nor the names of its
# contributors may be used to endorse or promote products derived from this
# software without specific prior written permission.
# 
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
##################################################################################



import unittest

import os
import sys

sys.path.append(os.path.join( '..', 'src' ))
from heuristicDivide import *

# TODO: add tests for all the other functions in biconnected that we haven't test here
class getComponentsTest(unittest.TestCase):

  def testGetComponents(self):
    """
    Test the conversion node->component mapping to component->nodes mapping.
    """
    componentsStructure = {}
    componentsStructure['1'] = 1
    componentsStructure['2'] = 1
    componentsStructure['3'] = 1
    componentsStructure['4'] = 2 
    componentsStructure['5'] = 2 
    componentsStructure['6'] = 3 
    componentsStructure['7'] = 4 

    expectedResult = {}
    expectedResult[1] = ['1', '2', '3']
    expectedResult[2] = ['4', '5']
    expectedResult[3] = ['6']
    expectedResult[4] = ['7']
    self.assertEqual(getComponents(componentsStructure), expectedResult)
    

class findSingletonsTest(unittest.TestCase):
  
  def testFindSingletons(self): 
    # test with several components and two singletons
    componentsStructure = {}
    componentsStructure['1'] = 1
    componentsStructure['2'] = 1
    componentsStructure['3'] = 1
    componentsStructure['4'] = 2 
    componentsStructure['5'] = 2 
    componentsStructure['6'] = 3 
    componentsStructure['7'] = 4 
    self.assertEqual(sorted(findSingletons(componentsStructure)), ['6', '7'])

    # test with one singleton component
    componentsStructure = {}
    componentsStructure['1'] = 1
    self.assertEqual(sorted(findSingletons(componentsStructure)), ['1'])


class findDisjointMaximalCyclesTest(unittest.TestCase):
  
  def testOneNode(self):
    self.gr = graph()
    # Add nodes and edges
    self.gr.add_nodes(['1'])

    self.assertEqual(findDisjointMaximalCycles(self.gr), [['1']])

  def testFewSingletonNodes(self):
    self.gr = graph()
    # Add nodes and edges
    self.gr.add_nodes(['1'])
    self.gr.add_nodes(['2'])
    self.gr.add_nodes(['3'])

    cycles = findDisjointMaximalCycles(self.gr)
    self.assertTrue(['1'] in cycles)
    self.assertTrue(['2'] in cycles)
    self.assertTrue(['3'] in cycles)
    self.assertEqual(len(cycles), 3)

  def testOneEdgeGraph(self):
    self.gr = graph()
    # Add nodes and edges
    self.gr.add_nodes(['1'])
    self.gr.add_nodes(['2'])
    self.gr.add_edge(('1', '2'))
    
    self.assertEqual(findDisjointMaximalCycles(self.gr), [['1', '2']])

  def testTwoEdgesGraph(self):
    self.gr = graph()
    # Add nodes and edges
    self.gr.add_nodes(['1'])
    self.gr.add_nodes(['2'])
    self.gr.add_nodes(['3'])
    self.gr.add_edge(('1', '2'))
    self.gr.add_edge(('2', '3'))
    
    cycles = findDisjointMaximalCycles(self.gr)
    self.assertTrue(
      (['1'] in cycles and ['2', '3'] in cycles) or
      (['3'] in cycles and ['1', '2'] in cycles)
    )
    self.assertEqual(len(cycles), 2)

  def testLargeCircle(self):
    self.gr = graph()
    # Add nodes and edges
    self.gr.add_nodes(['1', '2', '3', '4', '5', '6'])
    self.gr.add_edge(('1','2'), 1)
    self.gr.add_edge(('2','3'), 1)
    self.gr.add_edge(('3','1'), 1)
    self.gr.add_edge(('3','4'), 10)
    self.gr.add_edge(('4','5'), 1)
    self.gr.add_edge(('5','6'), 1)
    self.gr.add_edge(('6','4'), 1)
    self.gr.add_edge(('6','1'), 10)
    self.assertEqual(findDisjointMaximalCycles(self.gr), [['1', '2', '3', '4', '5', '6']])
    
  def testLargeCircleWithBridge(self):
    self.gr = graph()
    # Add nodes and edges
    self.gr.add_nodes(['1', '2', '3', '4', '5', '6'])
    self.gr.add_edge(('1','2'), 1)
    self.gr.add_edge(('2','3'), 1)
    self.gr.add_edge(('3','1'), 1)
    self.gr.add_edge(('3','4'), 10)
    self.gr.add_edge(('4','5'), 1)
    self.gr.add_edge(('5','6'), 1)
    self.gr.add_edge(('6','4'), 1)

    cycles = findDisjointMaximalCycles(self.gr)

    self.assertTrue(['1', '2', '3'] in cycles)
    self.assertTrue(['4', '5', '6'] in cycles)
    self.assertEqual(len(cycles), 2)


class cycleLengthTest(unittest.TestCase):

  def testUndirectedCycleLength(self):
    self.gr = graph()
    # Add nodes and edges
    self.gr.add_nodes(['1', '2', '3', '4', '5', '6'])
    self.gr.add_edge(('1','2'), 1)
    self.gr.add_edge(('2','3'), 1)
    self.gr.add_edge(('3','1'), 1)
    self.gr.add_edge(('3','4'), 10)
    self.gr.add_edge(('4','5'), 1)
    self.gr.add_edge(('5','6'), 1)
    self.gr.add_edge(('6','4'), 1)
    self.gr.add_edge(('6','1'), 10)

    self.assertEqual(cycleLength(['1', '2', '3'], self.gr), 3)
    self.assertEqual(cycleLength(['1', '2', '3', '4', '5', '6'], self.gr), 24)
    self.assertEqual(cycleLength(['4', '5', '6'], self.gr), 3) 
    
  def testDirectedCycleLength(self):
    self.gr = digraph()
    # Add nodes and edges
    self.gr.add_nodes(['1', '2', '3'])
    self.gr.add_edge(('1','2'), 1)
    self.gr.add_edge(('2','3'), 1)
    self.gr.add_edge(('3','1'), 1)

    self.gr.add_edge(('3','2'), 2)
    self.gr.add_edge(('2','1'), 2)
    self.gr.add_edge(('1','3'), 2)

    self.assertEqual(cycleLength(['1', '2', '3'], self.gr), 3)
    self.assertEqual(cycleLength(['3', '2', '1'], self.gr), 6)

  def testRaisesNonExistingEdge(self):
    self.gr = digraph()
    # Add nodes and edges
    self.gr.add_nodes(['1', '2', '3'])
    self.gr.add_edge(('1','2'), 1)
    self.gr.add_edge(('2','3'), 1)
    self.gr.add_edge(('3','1'), 1)
    self.assertRaises(Exception, cycleLength, ['3', '2', '1'], self.gr)


class assignKAgentsTest(unittest.TestCase):
  
  def testOneNode(self):
    self.gr = digraph()
    # Add nodes and edges
    self.gr.add_nodes(['1'])
    cycles = [['1']]
    # not possible with 0 agents..
    self.assertEqual( assignKAgents(self.gr, cycles, 0.0001, 0), None)
    # possible with 1 agents..
    self.assertEqual( assignKAgents(self.gr, cycles, 0.0001, 1), ([['1']], [0], [1]) )

  def testOneCycle(self):
    # Setup
    self.gr = digraph()
    # Add nodes and edges
    self.gr.add_nodes(['1', '2', '3'])
    self.gr.add_edge(('1','2'), 1)
    self.gr.add_edge(('2','3'), 1)
    self.gr.add_edge(('3','1'), 1)

    self.gr.add_edge(('2','1'), 2)
    self.gr.add_edge(('3','2'), 2)
    self.gr.add_edge(('1','3'), 2)

    # create a simple cycle and test it 
    # Note that assignKAgents() checks both directions of the cycle:
    # The cycle in the given direction would fail the below, 
    # but assignKAgents() check both directions
    cycle = ['3', '2', '1']
    cycles =  [cycle]

    # can satisfy idleness requirement by either cycling or guarding
    self.assertNotEqual( assignKAgents(self.gr, cycles, 1, 3), None) 

    # cannot satisfy idleness requirement
    self.assertEqual( assignKAgents(self.gr, cycles, 1, 2), None) 

    # cannot circle, but can still guard, cycle direction shouldn't change (as it's meaningless) 
    # TODO: ORDER MIGHT AFFECT THIS TEST
    self.assertEqual( assignKAgents(self.gr, cycles, 0.0001, 3), ([['3'], ['2'], ['1']], [0, 0, 0], [1,1,1])) 

    # can satisfy idleness by circling in the opposite direction
    self.assertEqual( assignKAgents(self.gr, cycles, 1.5001, 2), ([['1', '2', '3']], [1.5], [2])) 

  def testMoreThanOneCycle(self):
    # Setup
    self.gr = digraph()
    # Add nodes and edges
    self.gr.add_nodes(['1', '2', '3', '4', '5', '6'])
    self.gr.add_edge(('1','2'), 2)
    self.gr.add_edge(('2','3'), 2)
    self.gr.add_edge(('3','1'), 2)

    self.gr.add_edge(('2','1'), 2)
    self.gr.add_edge(('3','2'), 2)
    self.gr.add_edge(('1','3'), 2)


    self.gr.add_edge(('4','5'), 3)
    self.gr.add_edge(('5','6'), 3)
    self.gr.add_edge(('6','4'), 3)

    self.gr.add_edge(('5','4'), 3)
    self.gr.add_edge(('6','5'), 3)
    self.gr.add_edge(('4','6'), 3)

    cycles = [ ['3', '2', '1'], ['6', '5', '4'] ]
    # Test for different combinations of f (required freq) and k (#robots)
    # ==================================>>>           f      k  freq-list  #robots
    # TODO: ORDER MIGHT AFFECT THIS TEST
    self.assertEqual( assignKAgents(self.gr, cycles, 0.0001, 6), ([ ['3'], ['2'], ['1'], ['6'], ['5'], ['4']], [0, 0, 0, 0, 0, 0], [1, 1, 1, 1, 1, 1]) ) 
    self.assertEqual( assignKAgents(self.gr, cycles, 0.0001, 5), None ) 
    self.assertEqual( assignKAgents(self.gr, cycles, 2.9,    5), None ) 
    self.assertEqual( assignKAgents(self.gr, cycles, 3,      5), ([['3', '2', '1'], ['6'], ['5'], ['4']], [3.0, 0, 0, 0], [2.0, 1, 1, 1]) ) 
    self.assertEqual( assignKAgents(self.gr, cycles, 3,      4), None ) 
    self.assertEqual( assignKAgents(self.gr, cycles, 4.4,    4), None ) 
    self.assertEqual( assignKAgents(self.gr, cycles, 4.5,    4), ([ ['3', '2', '1'], ['6', '5', '4'] ], [3, 4.5], [2, 2]) ) 



class divideTo2CirclesTest(unittest.TestCase):

  def setUp(self):
    self.gr = graph()
    # Add nodes and edges
    self.gr.add_nodes(['1', '2', '3', '4', '5', '6'])
    self.gr.add_edge(('1','2'), 1)
    self.gr.add_edge(('2','3'), 1)
    self.gr.add_edge(('3','1'), 1)
    self.gr.add_edge(('3','4'), 10)
    self.gr.add_edge(('4','5'), 1)
    self.gr.add_edge(('5','6'), 1)
    self.gr.add_edge(('6','4'), 1)
    self.gr.add_edge(('6','1'), 10)

  def testDivideTo2Circles(self):
    # =============================>>>              f       k
    self.assertEqual( divideTo2Circles(self.gr,    0.001,   5), None) 
    self.assertNotEqual( divideTo2Circles(self.gr, 1.50001, 5), None) 
    self.assertEqual( divideTo2Circles(self.gr,    1.50001, 4), 
                     # <list of cycles, list of cycle-frequencies, list of robot assignments>
                     ( [ ['1', '2', '3'], ['4', '5', '6'] ], [1.5, 1.5], [2, 2] )
                     )
    self.assertEqual( divideTo2Circles(self.gr, 1.49999, 4), None) 


class inducedGraphTest(unittest.TestCase):

  def testSameGraph(self):
    self.gr = digraph()
    self.gr.add_nodes(['1', '2', '3', '4', '5', '6'])
    self.gr.add_edge(('1','2'), 200)
    self.gr.add_edge(('2','3'), 1)
    self.assertEqual(inducedGraph(self.gr, self.gr.nodes()), self.gr)

  def testSubGraph(self):
    self.gr = digraph()
    nodes = ['1', '2', '3', '4', '5', '6']
    self.gr.add_nodes(nodes)
    for i in range(len(nodes)): 
      for j in range(i + 1, len(nodes)):
        self.gr.add_edge((nodes[i], nodes[j]), 1)
        self.gr.add_edge((nodes[j], nodes[i]), 2)

    expected = digraph()
    nodes = ['2', '4', '5']
    expected.add_nodes(nodes)
    for i in range(len(nodes)): 
      for j in range(i + 1, len(nodes)):
        expected.add_edge((nodes[i], nodes[j]), 1)
        expected.add_edge((nodes[j], nodes[i]), 2)
    self.assertEqual(inducedGraph(self.gr, nodes), expected)



class heuristicDivideTest(unittest.TestCase):

###### DELETED BECAUSE heuristicDivide() SHOULD GET A DIRECTED GRAPH

#  def testUndirectedGraph(self):
#    print >> sys.stderr, 'START'
#    self.gr = graph()
#    # Add nodes and edges
#    self.gr.add_nodes(['1', '2', '3', '4', '5', '6'])
#    self.gr.add_edge(('1','2'), 1)
#    self.gr.add_edge(('2','3'), 1)
#    self.gr.add_edge(('3','1'), 1)
#    self.gr.add_edge(('3','4'), 10)
#    self.gr.add_edge(('4','5'), 1)
#    self.gr.add_edge(('5','6'), 1)
#    self.gr.add_edge(('6','4'), 1)
#    self.gr.add_edge(('6','1'), 10)
#    # TODO: works well, but currently there is a random tie-break
#    # add assertion and solve it
#    #=============================>>>              f       k
#    self.assertEqual( heuristicDivide(self.gr,    1.50001, 4), 
#                     # <list of cycles, list of cycle-frequencies, list of robot assignments>
#                     [ (['1', '2', '3'], 2), (['4', '5', '6'] , 2) ] 
#                     )

  def testDirectedGraph(self):
    self.gr = digraph()
    # Add nodes and edges
    self.gr.add_nodes(['1', '2', '3', '4', '5', '6'])
    self.gr.add_edge(('1','2'), 200)
    self.gr.add_edge(('2','1'), 1)
    self.gr.add_edge(('2','3'), 200)
    self.gr.add_edge(('3','2'), 1)
    self.gr.add_edge(('3','1'), 200)
    self.gr.add_edge(('1','3'), 1)

    self.gr.add_edge(('4','5'), 200)
    self.gr.add_edge(('5','4'), 1)
    self.gr.add_edge(('5','6'), 200)
    self.gr.add_edge(('6','5'), 1)
    self.gr.add_edge(('6','4'), 200)
    self.gr.add_edge(('4','6'), 1)

    self.gr.add_edge(('3','4'), 1000)
    self.gr.add_edge(('4','3'), 1000)
    self.gr.add_edge(('6','1'), 1000)
    self.gr.add_edge(('1','6'), 1000)
    # should figure out to that it should cycle in the opposite direction
    #=============================>>>              f       k
    self.assertEqual( heuristicDivide(self.gr,    1.50001, 4), 
                     # <list of cycles, list of cycle-frequencies, list of robot assignments>
                     [ (['3', '2', '1'], 1.5, 2), (['6', '5', '4'], 1.5, 2) ] 
                     )
    # test that even when we allow the frequency to be infinity, it still finds a good solution
    self.assertEqual( heuristicDivide(self.gr,    9999999, 4), 
                     # <list of cycles, list of cycle-frequencies, list of robot assignments>
                     [ (['3', '2', '1'], 1.5, 2), (['6', '5', '4'], 1.5 , 2) ] 
                     )

if __name__ == "__main__":
    unittest.main()

