##################################################################################
# Copyright (c) 2011, Daniel Urieli, Peter Stone
# University of Texas at Austin
# All right reserved
# 
# Based On:
# 
# Copyright (c) 2000-2003, Jelle Kok, University of Amsterdam
# All rights reserved.
# 
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
# 
# 1. Redistributions of source code must retain the above copyright notice, this
# list of conditions and the following disclaimer.
# 
# 2. Redistributions in binary form must reproduce the above copyright notice,
# this list of conditions and the following disclaimer in the documentation
# and/or other materials provided with the distribution.
# 
# 3. Neither the name of the University of Amsterdam nor the names of its
# contributors may be used to endorse or promote products derived from this
# software without specific prior written permission.
# 
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
##################################################################################




import unittest

import os
import sys
from math import *

sys.path.append(os.path.join( '..', 'src' ))
import shipModels
from rrtTree import *
import state
import actions
from perceptModels import *


class testRRTTreeXYOrientationEuclidDist(unittest.TestCase):
  """
  Tests the RRTTreeXYOrientationEuclidDist class. 
  We tests more basic functions first, so the order matters,
  for correct bottom-up testing, and error detection.
  """
  def setUp(self):
    self.shipState = state.ShipExternalState(0,0,0,0)
    self.tree = RRTTreeXYOrientationEuclidDist(self.shipState, 10, 5)

  def testAddNode(self):
    """Tests the addNode() method."""
    newState = state.ShipExternalState(10, 10, 10, 10)
    newNode = RRTNode(newState, self.tree.root)
    self.tree.addNode(newNode)

    self.assertEqual(len(self.tree.nodes), 2)
    self.assertTrue(newNode == self.tree.nodes[0] or
                    newNode == self.tree.nodes[1])

  def testGetPathTo(self):
    """Tests the getPathTo() method."""
    newState = state.ShipExternalState(10, 10, 10, 10)
    action = actions.CompositeAction()
    action.appendAction(actions.PrimitiveAction('somefunction',['somearg']))
    newNode = RRTNode(newState, self.tree.root, action)
    # Assuming here that the testAddNode() function passed
    self.tree.addNode(newNode)

    path = self.tree.getPathTo(newState)

    # path should be of length 3: 
    # 1. An empty action for the root
    # 2. A transition action to move to newState
    # 3. A stop action
    self.assertEqual(len(path), 3)

    # First action should be empty
    action = path[0]
    self.assertEqual(action.primitiveActions, [])

    # Second action should be the transition, 
    # which is just the delta currently
    action = path[1]
    self.assertTrue(len(action.primitiveActions) == 1)
    self.assertEqual(action.primitiveActions[0].functionName, 'somefunction')
    self.assertEqual(action.primitiveActions[0].args, ['somearg'])

    # Third action should be the stop action
    action = path[2]
    self.assertTrue(len(action.primitiveActions) == 1)
    self.assertEqual(action.primitiveActions[0].functionName, 'stop')
    self.assertEqual(action.primitiveActions[0].args, [])

  def testFindClosestNode(self):
    """Tests the findClosestNode() method"""
    newState = state.ShipExternalState(10, 10, 10, 10)
    newNode = RRTNode(newState, self.tree.root)
    # Assuming here that the testAddNode() function passed
    self.tree.addNode(newNode)

    testState = state.ShipExternalState(6, 6, 6, 6)
    node = self.tree.findClosestNode(testState)
    self.assertTrue(node.state.x == 10 and
                    node.state.y == 10 and
                    node.state.orientation == 10)

    testState = state.ShipExternalState(4, 4, 4, 4)
    node = self.tree.findClosestNode(testState)
    self.assertTrue(node.state.x == 0 and
                    node.state.y == 0 and
                    node.state.orientation == 0)


    




'''
class testRRT(unittest.TestCase):
  
  def setUp(self):
    self.shipState = state.ShipExternalState(0,0,0,0)
    self.rrt = agents.RRT(self.shipState)
  
  def testExtend(self):
    "Tests that extend adds a node on the way to a new state, within epsilon"
    sampledState = state.ShipExternalState(10, 10, 10, 10)
    newState = self.rrt.extend(sampledState)
    
    # There should be two nodes
    self.assertEqual(len(self.rrt.nodes), 2)

    # New state should be epsilon towards sampledState
    rrtEpsilon = self.rrt.epsilon
    expectedNewStateX = abs(sampledState.x - self.shipState.x) * rrtEpsilon
    expectedNewStateY = abs(sampledState.y - self.shipState.y) * rrtEpsilon
    expectedNewStateOrient = abs(sampledState.orientation - self.shipState.orientation) * rrtEpsilon

    delta = 0.0000001
    self.assertTrue( abs(newState.x - expectedNewStateX) < delta)
    self.assertTrue( abs(newState.y - expectedNewStateY) < delta)
    self.assertTrue( abs(newState.orientation - expectedNewStateOrient) < delta)

  def testAddNode(self):
    self.rrt.addNode(None)
    self.assertEqual(len(self.rrt.nodes), 2)

  def testGetPathTo(self):
    "Assuming testExtend() passed"
    sampledState = state.ShipExternalState(10, 10, 10, 10)
    newState = self.rrt.extend(sampledState)
    path = self.rrt.getPathTo(newState)
    
    # path should be of length 3: 
    # 1. An empty action for the root
    # 2. A transition action to move to newState
    # 3. A stop action
    self.assertEqual(len(path), 3)

    # First action should be empty
    action = path[0]
    self.assertEqual(action.primitiveActions, [])

    # Second action should be the transition, 
    # which is just the delta currently
    action = path[1]
    rrtEpsilon = self.rrt.epsilon
    expectedDiffX = abs(sampledState.x - self.shipState.x) * rrtEpsilon - self.shipState.x
    expectedDiffY = abs(sampledState.y - self.shipState.y) * rrtEpsilon - self.shipState.y
    expectedDiffOrient = abs(sampledState.orientation - self.shipState.orientation) * rrtEpsilon - self.shipState.orientation

    expectedSpeedCommand = sqrt(expectedDiffX * expectedDiffX + expectedDiffY * expectedDiffY)
    expectedOrientCommand = expectedDiffOrient

    delta = 0.0000001
    for a in action.primitiveActions: 
      cmdName = a.functionName
      self.assertTrue(cmdName == "setEngineSpeed" or cmdName == "setSteering")

      if cmdName == "setEngineSpeed":
        speed = a.args[0]
        self.assertTrue(abs(speed - expectedSpeedCommand) < delta)
      elif cmdName == "setSteering":
        orientation = a.args[0]
        self.assertTrue(abs(orientation - expectedDiffOrient) < delta)
        
'''




    


if __name__ == "__main__":
    unittest.main()
