##################################################################################
# Copyright (c) 2011, Daniel Urieli, Peter Stone
# University of Texas at Austin
# All right reserved
# 
# Based On:
# 
# Copyright (c) 2000-2003, Jelle Kok, University of Amsterdam
# All rights reserved.
# 
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
# 
# 1. Redistributions of source code must retain the above copyright notice, this
# list of conditions and the following disclaimer.
# 
# 2. Redistributions in binary form must reproduce the above copyright notice,
# this list of conditions and the following disclaimer in the documentation
# and/or other materials provided with the distribution.
# 
# 3. Neither the name of the University of Amsterdam nor the names of its
# contributors may be used to endorse or promote products derived from this
# software without specific prior written permission.
# 
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
##################################################################################




import unittest

import os
import sys

sys.path.append(os.path.join( '..', 'src' ))
import shipModels
import actions
from perceptModels import *



class PerceptionModulesTest(unittest.TestCase):
  "Test different perception modules"
  
  def testCompleteStatePerceptionModule(self):
    "Tests that CompleteStatePerceptionModule percepts the full state"
    for state in [ 'bla', 1, (1,2) ]:
      p = shipModels.CompleteStatePerceptionModule()
      perceptsList = p.getListOfPercepts(state)
      self.assertEqual(perceptsList.percepts[0].state, state)




class ShipInterfaceFunctions(unittest.TestCase):


  def testGetPercepts(self):
    "getPercepts() is an interface function that returns a PerceptList"
    s = shipModels.Ship()
    state = None
    self.assertTrue(isinstance(s.getPercepts(state), PerceptList))


  def testExecuteAction(self):
    "test executeAction works with different actions on ship"

    # list of 2 standard actions to execute
    s = shipModels.Ship()
    action = actions.CompositeAction()
    action.appendAction(actions.PrimitiveAction("setEngineSpeed", [5]))
    action.appendAction(actions.PrimitiveAction("setSteering", [10]))
    s.executeAction(action)
    self.assertEqual(s.engineSpeed, 5)
    self.assertEqual(s.steering, 10)

    # list of 1 action to execute
    s = shipModels.Ship()
    action = actions.CompositeAction()
    action.appendAction(actions.PrimitiveAction("setEngineSpeed", [7]))
    s.executeAction(action)
    self.assertEqual(s.engineSpeed, 7)

    # list of 0 actions to execute
    s = shipModels.Ship()
    engineSpeed = s.engineSpeed
    steering = s.steering
    action = actions.CompositeAction()
    s.executeAction(action)
    self.assertEqual(s.engineSpeed, engineSpeed)
    self.assertEqual(s.steering, steering)


#class ShipBadInput(unittest.TestCase):
#  def testWrongComposedAction(self):
#    # wrong pairs, function that doesn't exist and so on
#
#

  # actions
  def testSteering(self):
    "Test the setSteering() function"
    origSpeed = 15
    origSteering = 32
    maxSpeedLimit = 30
    minSpeedLimit = -10
    maxSteeringLimit = 45
    minSteeringLimit = -45

    s = shipModels.Ship(initialEngineSpeed = origSpeed,
                         initialSteering = origSteering,
                         maxSpeed = maxSpeedLimit, 
                         minSpeed = minSpeedLimit, 
                         maxSteering = maxSteeringLimit,
                         minSteering = minSteeringLimit)


    s.setSteering(3)
    self.assertEqual(s.steering, 3)

    s.setSteering(-3)
    self.assertEqual(s.steering, -3)

    s.setSteering(0)
    self.assertEqual(s.steering, 0)

    # setting the limit speeds should work
    s.setSteering(maxSteeringLimit)
    self.assertEqual(s.steering, maxSteeringLimit) 

    # setting the limit speeds should work
    s.setSteering(minSteeringLimit)
    self.assertEqual(s.steering, minSteeringLimit) 


  def testSpeed(self):
    "Test the setEngineSpeed() function"
    origSpeed = 15
    origSteering = 32
    maxSpeedLimit = 30
    minSpeedLimit = -10
    maxSteeringLimit = 45
    minSteeringLimit = -45

    s = shipModels.Ship(initialEngineSpeed = origSpeed,
                         initialSteering = origSteering,
                         maxSpeed = maxSpeedLimit, 
                         minSpeed = minSpeedLimit, 
                         maxSteering = maxSteeringLimit,
                         minSteering = minSteeringLimit)


    s.setEngineSpeed(3)
    self.assertEqual(s.engineSpeed, 3)

    s.setEngineSpeed(-3)
    self.assertEqual(s.engineSpeed, -3)

    s.setEngineSpeed(0)
    self.assertEqual(s.engineSpeed, 0)

    # setting the limit speeds should work
    s.setEngineSpeed(maxSpeedLimit)
    self.assertEqual(s.engineSpeed, maxSpeedLimit) 

    # setting the limit speeds should work
    s.setEngineSpeed(minSpeedLimit)
    self.assertEqual(s.engineSpeed, minSpeedLimit) 



class ShipBadInput(unittest.TestCase):
#  
#  def testWrongTypePercept(self):
#    # test that percept is of type CompleteStatePercept
#
  def setUp(self):
    self.origSpeed = 15
    self.origSteering = 32
    self.maxSpeedLimit = 30
    self.minSpeedLimit = -10
    self.maxSteeringLimit = 45
    self.minSteeringLimit = -45

    self.s = shipModels.Ship(initialEngineSpeed = self.origSpeed,
                         initialSteering = self.origSteering,
                         maxSpeed = self.maxSpeedLimit, 
                         minSpeed = self.minSpeedLimit, 
                         maxSteering = self.maxSteeringLimit,
                         minSteering = self.minSteeringLimit)
    

  def testTooLargeSpeed(self):
    "Should ignore settings of too large speed"


    # no change should be made
    self.s.setEngineSpeed(self.maxSpeedLimit + 1)
    self.assertEqual(self.s.engineSpeed, self.origSpeed) 


  def testTooSmallSpeed(self):
    "Should ignore settings of too small speed"

    # no change should be made
    self.s.setEngineSpeed(self.minSpeedLimit - 1)
    self.assertEqual(self.s.engineSpeed, self.origSpeed) 


  def testTooLargeSteering(self):
    "Should ignore settings of too large steering"

    # no change should be made
    self.s.setSteering(self.maxSteeringLimit + 1)
    self.assertEqual(self.s.steering, self.origSteering) 


  def testTooSmallSteering(self):
    "Should ignore settings of too small steering"

    # no change should be made
    self.s.setSteering(self.minSteeringLimit - 1)
    self.assertEqual(self.s.steering, self.origSteering) 

    


if __name__ == "__main__":
    unittest.main()
